"""A graphical application to interact with your power usage statistics."""
from __future__ import annotations

import asyncio
from tkinter import PhotoImage
from typing import TYPE_CHECKING

import customtkinter as ctk
from CTkMessagebox import CTkMessagebox

from power_comparison.plan_comparison_screen import PlanComparisonScreen
from power_comparison.usage_view_screen import UsageViewScreen

if TYPE_CHECKING:
    from power_comparison.controller import Controller


class View:
    """A graphical application to interact with your power usage statistics."""

    _root: ctk.CTk
    _controller: Controller
    _close: bool = False

    def __init__(self, controller: Controller) -> None:
        """Initialize the App."""
        self._controller = controller
        self._root = ctk.CTk()
        self._root.minsize(width=1280, height=720)
        self._root.rowconfigure(0, weight=1)
        self._root.columnconfigure(0, weight=1)
        self._root.protocol("WM_DELETE_WINDOW", self.close_view)
        icon = PhotoImage(file=self.get_controller().get_icon_path())
        self._root.iconphoto(True, icon)
        self.launch_login_screen()
        asyncio.get_event_loop().run_until_complete(self.exec(1 / 20))

    def close_view(self) -> None:
        """Close view."""
        self._close = True

    async def exec(self, interval: float) -> None:
        """Execute view."""
        while True:
            if self._close:
                self._root.destroy()
                break
            self._root.update()
            await asyncio.sleep(interval)

    def config_grid(
        self, frame: ctk.CTkFrame, rows: list[int], columns: list[int]
    ) -> None:
        """Configure the amount of rows and columns on the root."""
        for i, weight in enumerate(rows):
            frame.rowconfigure(i, weight=weight)
        for i, weight in enumerate(columns):
            frame.columnconfigure(i, weight=weight)

    def set_title(self, title: str) -> None:
        """Configure the window title."""
        self._root.title(title)

    def clear_screen(self) -> None:
        """Clear the screen."""
        for widget in self._root.winfo_children():
            widget.destroy()

    def new_frame(self) -> ctk.CTkFrame:
        """Return a new frame.

        This returns a new 'root'-like frame for a new window.
        It also clears the screen, and configures this frame to have
        one row and column.
        """
        self.clear_screen()
        frame = ctk.CTkFrame(self._root)
        frame.grid(row=0, column=0, sticky="NEWS")
        self.config_grid(frame, [1], [1])
        return frame

    def set_padding(self, frame: ctk.CTkFrame, padx: int, pady: int) -> None:
        """Set the padding of all children in a frame."""
        for child in frame.winfo_children():
            child.grid(padx=padx, pady=pady)

    def get_controller(self) -> Controller:
        """Return the controller."""
        return self._controller

    def launch_login_screen(self) -> None:
        """Launch the login screen."""
        LoginScreen(self)

    def launch_data_download(self) -> None:
        """Launch data download screen."""
        DataDownloadScreen(self)

    def launch_main_screen(self) -> None:
        """Launch main screen."""
        MainScreen(self)

    def launch_usage_view_screen(self) -> None:
        """Launch usage view screen."""
        UsageViewScreen(self)

    def launch_plan_comparison_screen(self) -> None:
        """Launch plan comparison screen."""
        PlanComparisonScreen(self)


class LoginScreen:
    """Define the Login or Registration selection screen."""

    _app: View
    _selected_connector: ctk.CTkOptionMenu
    _username: ctk.StringVar
    _password: ctk.StringVar

    def __init__(self, app: View) -> None:
        """Create LoginScreen."""
        self._app = app
        self.tk_init()

    def tk_init(self) -> None:
        """Initialize Tkinter for this screen."""
        self._app.set_title("Power Comparison: Login")
        window_root = self._app.new_frame()
        frame = ctk.CTkFrame(window_root)
        frame.grid()
        self._app.config_grid(frame, [1, 1, 1, 1], [1, 2])
        ctk.CTkLabel(frame, text="Power Company:").grid(
            row=0, column=0, sticky="E"
        )
        self._selected_connector = ctk.CTkOptionMenu(
            frame,
            values=self._app.get_controller().get_connector_names(),
        )
        self._selected_connector.grid(row=0, column=1)
        ctk.CTkLabel(frame, text="Username/Email:").grid(
            row=1, column=0, sticky="E"
        )
        self._username = ctk.StringVar()
        ctk.CTkEntry(frame, textvariable=self._username).grid(row=1, column=1)
        ctk.CTkLabel(frame, text="Password:").grid(row=2, column=0, sticky="E")
        self._password = ctk.StringVar()
        password_entry = ctk.CTkEntry(
            frame, show="•", textvariable=self._password
        )
        password_entry.grid(row=2, column=1)
        password_entry.bind(
            "<Return>", lambda _: asyncio.create_task(self.next_clicked())
        )
        self._next_button = ctk.CTkButton(
            frame,
            text="Login",
            command=lambda: asyncio.create_task(self.next_clicked()),
        )
        self._next_button.grid(row=3, column=1, sticky="E")
        self._app.set_padding(frame, 5, 5)

    async def next_clicked(self) -> None:
        """Event handler for the next button being clicked."""
        self._next_button.configure(state="disabled")
        result = await self._app.get_controller().try_connect(
            self._selected_connector.get(),
            self._username.get(),
            self._password.get(),
        )
        if result is not None:
            CTkMessagebox(title=result[0], message=result[1], icon="cancel")
            self._next_button.configure(state="normal")
            return
        self._app.launch_data_download()


class DataDownloadScreen:
    """Define the Data Download screen."""

    _app: View
    _message: ctk.StringVar

    def __init__(self, app: View) -> None:
        """Create DataDownloadScreen."""
        self._app = app
        self.tk_init()
        self._app.get_controller().download_data(
            self._app.launch_main_screen, self.update_message
        )

    def tk_init(self) -> None:
        """Initialize Tkinter for this screen."""
        self._app.set_title("Power Comparison: Downloading Data")
        window_root = self._app.new_frame()
        frame = ctk.CTkFrame(window_root)
        frame.grid()
        self._app.config_grid(frame, [1], [1])
        self._message = ctk.StringVar()
        self._message.set("Downloading data")
        ctk.CTkLabel(frame, textvariable=self._message).grid(row=0, column=0)
        self._app.set_padding(frame, 5, 5)

    def update_message(self, message: str) -> None:
        """Update message for user feedback on download."""
        self._message.set(message)


class MainScreen:
    """Define the main option selection/dashboard screen."""

    _app: View

    def __init__(self, app: View) -> None:
        """Create MainScreen."""
        self._app = app
        self.tk_init()

    def tk_init(self) -> None:
        """Initialize Tkinter for this screen."""
        self._app.set_title("Power Comparison: Dashboard")
        window_root = self._app.new_frame()
        frame = ctk.CTkFrame(window_root)
        frame.grid()
        self._app.config_grid(frame, [1, 1], [1, 1])
        ctk.CTkButton(
            frame,
            text="Usage Data",
            command=self._app.launch_usage_view_screen,
        ).grid(row=0, column=0)
        ctk.CTkButton(
            frame,
            text="Plan Comparison",
            command=self._app.launch_plan_comparison_screen,
        ).grid(row=0, column=1)
        ctk.CTkButton(frame, text="Exit", command=self._app.close_view).grid(
            row=1, column=1
        )
        ctk.CTkButton(
            frame, text="Logout", command=self._app.launch_login_screen
        ).grid(row=1, column=0)
        self._app.set_padding(frame, 5, 5)



"""Define the plan comparison screen."""

from __future__ import annotations

from typing import TYPE_CHECKING

import customtkinter as ctk
from CTkMessagebox import CTkMessagebox
from customtkinter import StringVar
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure

if TYPE_CHECKING:
    from power_comparison.view import View


class PlanComparisonScreen:
    """Define the plan comparison and profile selection screen."""

    _app: View
    _selected_plan_set: ctk.CTkOptionMenu
    _legend_frame: ctk.CTkScrollableFrame
    _figure: Figure
    _canvas: FigureCanvasTkAgg
    _start_date: StringVar
    _end_date: StringVar

    def __init__(self, app: View) -> None:
        """Create PlanComparisonScreen."""
        self._app = app
        self.tk_init()

    def tk_init(self) -> None:
        """Initialize Tkinter for this screen."""
        self._app.set_title("Power Comparison: Plan Comparison")
        window_root = self._app.new_frame()
        self._app.config_grid(window_root, [1], [1, 4])
        left_frame = ctk.CTkFrame(window_root)
        left_frame.grid(row=0, column=0, sticky="NEWS")
        self._app.config_grid(left_frame, [1, 2], [1])
        frame = ctk.CTkFrame(left_frame)
        frame.grid(row=0, column=0)
        self._app.config_grid(frame, [1, 1, 1, 1], [1, 1])
        # Plan Selection
        ctk.CTkLabel(frame, text="Select group of plans:").grid(
            row=0, column=0, sticky="E"
        )
        self._selected_plan_set = ctk.CTkOptionMenu(
            frame, values=self._app.get_controller().get_profile_set_names()
        )
        self._selected_plan_set.grid(row=0, column=1)
        # Date Selection
        self._start_date = StringVar(
            value=self._app.get_controller().get_start_date().strftime("%x")
        )
        self._end_date = StringVar(
            value=self._app.get_controller().get_last_date().strftime("%x")
        )
        ctk.CTkLabel(frame, text="Start date:").grid(
            row=1, column=0, sticky="E"
        )
        ctk.CTkLabel(frame, text="End date:").grid(row=2, column=0, sticky="E")
        ctk.CTkEntry(frame, textvariable=self._start_date).grid(
            row=1, column=1
        )
        ctk.CTkEntry(frame, textvariable=self._end_date).grid(row=2, column=1)
        ctk.CTkButton(frame, text="Compare", command=self.update_plot).grid(
            row=3, column=0, columnspan=2
        )
        self._app.set_padding(frame, 5, 5)
        # Graph
        graph_frame = ctk.CTkFrame(window_root)
        graph_frame.grid(row=0, column=1, sticky="NESW")
        self._app.config_grid(graph_frame, [1], [1])
        self.setup_plot(graph_frame)
        # Legend
        self._legend_frame = ctk.CTkScrollableFrame(left_frame)
        self._legend_frame.grid(row=1, column=0, sticky="NESW")
        self._legend_frame.bind_all(
            "<Button-4>",
            lambda _: self._legend_frame._parent_canvas.yview(
                "scroll", -1, "units"
            ),
        )
        self._legend_frame.bind_all(
            "<Button-5>",
            lambda _: self._legend_frame._parent_canvas.yview(
                "scroll", 1, "units"
            ),
        )
        # Back Button
        back_frame = ctk.CTkFrame(window_root)
        back_frame.grid(row=0, column=0, sticky="NW")
        ctk.CTkButton(
            back_frame, text="Back", command=self._app.launch_main_screen
        ).grid()
        self._app.set_padding(back_frame, 5, 5)

    def update_plot(self) -> None:
        """Event handler for compare being clicked."""
        result = self._app.get_controller().get_comparison_data(
            self._selected_plan_set.get(),
            self._start_date.get(),
            self._end_date.get(),
        )
        if isinstance(result, tuple):
            CTkMessagebox(title=result[0], message=result[1], icon="cancel")
            return
        self._figure.clear()
        axes = self._figure.add_subplot()
        y_axis = [str(i) for i in range(1, len(result) + 1)]
        x_axis = [p[1] for p in result]
        axes.set_title(
            f"Comparison of power plans for {self._selected_plan_set.get()}"
        )
        minor_x_ticks = range(0, int(x_axis[-1]), 100)
        major_x_ticks = range(0, int(x_axis[-1]), 200)
        axes.set_xlabel("Estimated cost of plan in a year ($)")
        axes.set_xticks(major_x_ticks)
        axes.tick_params(axis="x", labelrotation=45)
        axes.set_xticks(minor_x_ticks, minor=True)
        axes.grid(visible=True, which="both", axis="x")
        axes.grid(which="minor", alpha=0.3)
        axes.set_ylabel("Power Plan")
        axes.yaxis.set_inverted(True)
        axes.set_yticks(range(len(y_axis)), labels=y_axis)
        axes.set_yticklabels(y_axis)
        axes.barh(y_axis, x_axis)
        self._canvas.draw()
        self.update_legend(result)

    def update_legend(self, info: list[tuple[str, float]]) -> None:
        """Update the legend with power plan information.

        Args:
            info: A list of power plan names, and estimated prices.
        """
        for widget in self._legend_frame.winfo_children():
            widget.destroy()
        for i, (name, value) in enumerate(info):
            ctk.CTkLabel(
                self._legend_frame,
                text=f"#{i + 1:<2} {name}",
            ).grid(row=i, column=0, sticky="W")
            ctk.CTkLabel(self._legend_frame, text=f"${value:.2f}").grid(
                row=i, column=1, sticky="E", padx=5
            )

    def setup_plot(self, frame: ctk.CTkFrame) -> None:
        """Setup comparison plot."""
        self._figure = Figure()
        self._canvas = FigureCanvasTkAgg(self._figure, frame)
        self._figure.subplots_adjust(left=0.2)
        self._canvas.get_tk_widget().grid(row=0, column=0, sticky="NESW")



"""Define the abstract class Connector."""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Self

if TYPE_CHECKING:
    from collections.abc import Callable
    from datetime import date


class AuthException(Exception):
    """Error to indicate we cannot authenticate to the API."""


class Connector(ABC):
    """The abstract class for an API Connector."""

    @classmethod
    @abstractmethod
    async def create(
        cls, username: str, password: str, timeout: int = 60
    ) -> Self:
        """Initialize and connect the Connector to its API.

        Args:
            username:
                Username of user.
            password:
                Password of user.
            timeout:
                Time to wait until giving up.

        Raises:
            AuthException when there is an error authenticating.
            asyncio.TimeoutError
        """

    @abstractmethod
    async def retrieve_usage(
        self,
        start_date: date | None = None,
        end_date: date | None = None,
        callback: Callable[[int], None] | None = None,
    ) -> list[tuple[date, list[float]]]:
        """Retrieve usage data from the Connector's API.

        Args:
            start_date:
                Default value is end_date - 365 days. Inclusive.
            end_date:
                Default value is today. Inclusive.
            callback:
                Default value is None. Callback accepts date ordinal for user
                feedback while downloading data.

        Returns:
            A list of dates with corresponding usage values.

        Raises:
            asyncio.TimeoutError
            AuthException when token becomes stale.
        """

    @staticmethod
    @abstractmethod
    def get_name() -> str:
        """Return the name of the power utility this connector connects to."""



"""A graphical application to interact with your power usage statistics."""
from __future__ import annotations

import asyncio
from datetime import date, datetime, timedelta
from typing import TYPE_CHECKING

from power_comparison.connectors import Connectors
from power_comparison.connectors.connector import AuthException
from power_comparison.default_values_utility import DefaultValuesUtility as DVU

if TYPE_CHECKING:
    from collections.abc import Callable

    from power_comparison.connectors.connector import Connector
    from power_comparison.data import Data, Profiles


class Controller:
    """A class to control the application."""

    _connector: Connector | None = None
    _data: Data
    _profiles: Profiles
    _callback: Callable[[str], None] | None = None
    _username: str | None = None

    def __init__(self, data: Data, profiles: Profiles) -> None:
        """Initialize the controller."""
        self._data = data
        self._profiles = profiles

    def get_icon_path(self) -> str:
        """Return the path to the app's icon."""
        return DVU.get_icon_png_path()

    def get_connector_names(self) -> list[str]:
        """Return the names of the connectors."""
        return list(Connectors.get_names().keys())

    def get_profile_set_names(self) -> list[str]:
        """Return the names of the power plan profile sets."""
        return self._profiles.get_profile_set_names()

    async def try_connect(
        self, connector_name: str, username: str, password: str
    ) -> None | tuple[str, str]:
        """Try connecting and authenticating with a connector.

        Return None, or an error title and message.
        """
        if connector_name == "":
            return (
                "No Power Utility Selected",
                "You haven't selected a power utility to connect to.",
            )
        if connector_name not in Connectors.get_names():
            return (
                "Invalid Power Utility Selected",
                "You haven't selected a valid power utility to connect to.",
            )
        if username.strip() == "":
            return (
                "No username/email",
                "You haven't entered a username/email",
            )
        if password.strip() == "":
            return "No password", "You haven't entered a password."
        try:
            self._connector = await Connectors.get_names()[
                connector_name
            ].create(username, password)
        except AuthException:
            return (
                "Invalid login",
                "Your username/email and/or your password are incorrect.",
            )
        except asyncio.TimeoutError:
            return (
                "Timed out",
                "Timed out trying to connect, \
make sure your internet is working.",
            )
        except ValueError:
            return (
                "Error retrieving info",
                "We encountered an error trying to retrieve your information, \
please try again later.",
            )
        self._data.initialize_user(username)
        return None

    def download_data(
        self,
        finished_callback: Callable[[], None],
        callback: Callable[[str], None] | None = None,
    ) -> None:
        """Download user data."""
        self._callback = callback
        asyncio.get_event_loop().create_task(
            self.data_download_call(finished_callback)
        )

    async def data_download_call(
        self, finished_callback: Callable[[], None]
    ) -> None:
        """Call and await's connectors retrieve usage."""
        if self._connector is None:
            msg = "Controller._connector not set"
            raise ValueError(msg)
        start_date = self._data.get_last_date()
        if start_date:
            start_date += timedelta(days=1)
        try:
            data = await self._connector.retrieve_usage(
                start_date=start_date,
                callback=self.user_feedback_callback,
            )
        except asyncio.TimeoutError:
            if self._callback:
                self._callback("Error: Downloading data timed out")
            return
        if self._callback is not None:
            self._callback("Saving downloaded data")
        self._data.ingest_data(data)
        finished_callback()

    def user_feedback_callback(self, date_ordinal: int) -> None:
        """Accept a date ordinal to callback stored callback with str."""
        if self._callback is None:
            return
        display_date = date.fromordinal(date_ordinal).strftime("%Y-%m-%d")
        self._callback(f"Retrieving usage data for date: {display_date}")

    def get_last_date(self) -> date:
        """Return default end date."""
        result = self._data.get_last_date()
        return date.today() if result is None else result

    def get_start_date(self) -> date:
        """Return default start date."""
        last_date = self.get_last_date()
        return (date.today() if last_date is None else last_date) - timedelta(
            days=365
        )

    def get_usage_data(
        self, start_date: str, end_date: str
    ) -> list[float] | tuple[str, str]:
        """Return Usage Data or Error title and message."""
        try:
            start = datetime.strptime(start_date, "%x").date()
            end = datetime.strptime(end_date, "%x").date()
        except ValueError:
            return (
                "Error parsing dates",
                f"Your dates must be in the format: {date.today().strftime('%x')}",
            )
        result = self._data.get_usage_per_hour(start, end)
        if result is None:
            return "No Data", "Error no data was found for this range."
        return result

    def get_comparison_data(
        self, plan_set_name: str, start_date: str, end_date: str
    ) -> list[tuple[str, float]] | tuple[str, str]:
        """Show comparison data in matplotlib display.

        Returns None on success or error messages on failure.
        """
        if plan_set_name == "":
            return (
                "No Profile Set Selected",
                "You haven't selected a set of plans to compare.",
            )
        if plan_set_name not in self._profiles.get_profile_set_names():
            return (
                "Invalid Profile Set Selected",
                "You haven't selected a valid set of plans to compare.",
            )
        try:
            start = datetime.strptime(start_date, "%x").date()
            end = datetime.strptime(end_date, "%x").date()
        except ValueError:
            return (
                "Error parsing dates",
                f"Your dates must be in the format: {date.today().strftime('%x')}",
            )
        usage_data = self._data.get_average_usage(start, end)
        if usage_data is None:
            return "No Data", "Error no data was found for this range."
        data = self._profiles.generate_plan_comparison(
            usage_data, plan_set_name
        )
        if data is None:
            return (
                "Error Fetching Profile Set",
                "We encountered an error fetching this profile set, \
and it is not available for comparison at this time.",
            )
        return data



"""Define the usage view screen."""
from __future__ import annotations

from typing import TYPE_CHECKING

import customtkinter as ctk
from CTkMessagebox import CTkMessagebox
from customtkinter import StringVar
from matplotlib.backends.backend_tkagg import (
    FigureCanvasTkAgg,
)
from matplotlib.figure import Figure

if TYPE_CHECKING:
    from matplotlib.axes import Axes
    from matplotlib.container import BarContainer

    from power_comparison.view import View


class UsageViewScreen:
    """Define the usage view screen."""

    _app: View
    _plot: BarContainer
    _axes: Axes
    _canvas: FigureCanvasTkAgg
    _start_date: StringVar
    _end_date: StringVar

    def __init__(self, app: View) -> None:
        """Create UsageViewScreen."""
        self._app = app
        self.tk_init()

    def tk_init(self) -> None:
        """Initialize Tkinter for this screen."""
        self._app.set_title("Power Comparison: Usage View")
        window_root = self._app.new_frame()
        self._app.config_grid(window_root, [1], [1, 4])
        frame = ctk.CTkFrame(window_root)
        frame.grid(row=0, column=0)
        self._app.config_grid(frame, [1, 1, 1], [1, 1])
        self._start_date = StringVar(
            value=self._app.get_controller().get_start_date().strftime("%x")
        )
        self._end_date = StringVar(
            value=self._app.get_controller().get_last_date().strftime("%x")
        )
        ctk.CTkLabel(frame, text="Start date:").grid(
            row=0, column=0, sticky="E"
        )
        ctk.CTkLabel(frame, text="End date:").grid(row=1, column=0, sticky="E")
        ctk.CTkEntry(frame, textvariable=self._start_date).grid(
            row=0, column=1
        )
        ctk.CTkEntry(frame, textvariable=self._end_date).grid(row=1, column=1)
        ctk.CTkButton(frame, text="Plot", command=self.update_plot).grid(
            row=2, column=0, columnspan=2
        )
        self._app.set_padding(frame, 5, 5)
        # Graph
        graph_frame = ctk.CTkFrame(window_root)
        graph_frame.grid(row=0, column=1)
        self.draw_plot(graph_frame)
        # Back Button
        back_frame = ctk.CTkFrame(window_root)
        back_frame.grid(row=0, column=0, sticky="NW")
        ctk.CTkButton(
            back_frame, text="Back", command=self._app.launch_main_screen
        ).grid()
        self._app.set_padding(back_frame, 5, 5)

    def update_plot(self) -> None:
        """Update usage data plot."""
        usage_data = self._app.get_controller().get_usage_data(
            self._start_date.get(), self._end_date.get()
        )
        if isinstance(usage_data, tuple):
            CTkMessagebox(
                title=usage_data[0], message=usage_data[1], icon="cancel"
            )
            return
        for count, rect in zip(usage_data, self._plot.patches):
            rect.set_height(count)
        self._axes.relim()
        self._axes.autoscale_view()
        self._canvas.draw()

    def draw_plot(self, frame: ctk.CTkFrame) -> None:
        """Draw usage data plot."""
        figure = Figure(dpi=100)
        self._canvas = FigureCanvasTkAgg(figure, frame)
        self._axes = figure.add_subplot()
        x_axis = range(24)
        self._axes.set_xticks(x_axis)
        self._axes.set_title("Average Power Usage Per Hour")
        self._axes.set_xlabel("Hour of day")
        self._axes.set_ylabel("Power Usage (KWh)")
        self._axes.grid(visible=True, which="both", axis="y")
        self._plot = self._axes.bar(x_axis, 24 * [0])
        self._canvas.get_tk_widget().grid(row=0, column=0)
        self.update_plot()



"""Define an Enum of Connectors"""

from enum import Enum
from .connector import Connector
from .contact_energy_connector import ContactEnergyConnector


class Connectors(Enum):
    """An Enum of Connectors."""

    CONTACT_ENERGY = ContactEnergyConnector

    @staticmethod
    def get_names() -> dict[str, Connector]:
        """Return a map of names of connectors with connectors."""
        return {
            connector.value.get_name(): connector.value
            for connector in Connectors
        }



"""Define the default values utility class."""

from pathlib import Path

import platformdirs


class DefaultValuesUtility:
    """Define utility functions for default values."""

    _PROFILES_DIR = "profiles"
    _DB_FILE_PATH = "data/user_data.db"
    _APP_NAME = "Power Comparison"
    _ICON_ICO_PATH = "img/power_compare.ico"
    _ICON_PNG_PATH = "img/power_compare.png"

    @staticmethod
    def create_dirs(filepath: str) -> None:
        """Ensure all directories in filepath are created."""
        Path(filepath).parent.mkdir(parents=True, exist_ok=True)

    @staticmethod
    def get_profiles_dir() -> str:
        """Return a path to the profiles directory."""
        return str(Path(__file__).parent / DefaultValuesUtility._PROFILES_DIR)

    @staticmethod
    def get_icon_ico_path() -> str:
        """Return a path to the program's icon."""
        return str(Path(__file__).parent / DefaultValuesUtility._ICON_ICO_PATH)

    @staticmethod
    def get_icon_png_path() -> str:
        """Return a path to the program's icon."""
        return str(Path(__file__).parent / DefaultValuesUtility._ICON_PNG_PATH)

    @staticmethod
    def get_db_file_path() -> str:
        """Return a path to the database file."""
        return str(
            Path(
                platformdirs.user_data_dir(
                    DefaultValuesUtility._APP_NAME, roaming=True
                )
            )
            / DefaultValuesUtility._DB_FILE_PATH
        )



"""Implement Connector for the Contact Energy API."""
from __future__ import annotations

from collections.abc import Callable
from datetime import date, timedelta
from typing import Self

import async_timeout
import contact_energy_nz
from contact_energy_nz import ContactEnergyApi, UsageDatum

from power_comparison.connectors import connector
from power_comparison.connectors.connector import Connector


class ContactEnergyConnector(Connector):
    """Implement Connector for the Contact Energy API."""

    _token: str
    _connector: ContactEnergyApi
    _timeout: int
    _UTILITY_NAME = "Contact Energy"

    @classmethod
    async def create(
        cls, username: str, password: str, timeout: int = 60
    ) -> Self:
        """Initialize the ContactEnergyConnector.

        Raises:
            AuthException:
                username and/or password incorrect.
            ValueError:
                Something else has gone wrong.
        """
        self = cls()
        self._timeout = timeout
        try:
            self._connector = await self._authenticate(username, password)
        except contact_energy_nz.AuthException as e:
            raise connector.AuthException from e
        self._token = self._connector.token
        await self._connector.account_summary()
        return self

    async def _authenticate(
        self, username: str, password: str
    ) -> ContactEnergyApi:
        """Authenticate with the Contact API."""
        async with async_timeout.timeout(self._timeout):
            return await ContactEnergyApi.from_credentials(username, password)

    async def retrieve_usage(
        self,
        start_date: date | None = None,
        end_date: date | None = None,
        callback: Callable[[int], None] | None = None,
    ) -> list[tuple[date, list[float]]]:
        """Retrieve usage data from the Connector's API.

        Arguments:
            start_date:
                Default value is end_date - 365 days. Inclusive.
            end_date:
                Default value is today. Inclusive.
            callback:
                Default value is None. Callback accepts date ordinal for user
                feedback while downloading data.

        Throws:
            asyncio.TimeoutError
        """
        end_date = end_date if end_date else date.today()
        start_date = (
            start_date if start_date else end_date - timedelta(days=365)
        )
        data: list[list[UsageDatum]] = []
        valid_data = False
        day = end_date
        while day >= start_date:
            if callback:
                callback(day.toordinal())
            async with async_timeout.timeout(self._timeout):
                day_data = await self._connector.get_hourly_usage(day)
                if day_data is None or len(day_data) == 0:
                    if valid_data:
                        break
                else:
                    data.append(day_data)
                    valid_data = True
            day -= timedelta(days=1)
        return [
            (
                usage_datums[0].date.date(),
                [usage_datum.value for usage_datum in usage_datums],
            )
            for usage_datums in data
        ]

    @staticmethod
    def get_name() -> str:
        """Return the name of the power utility this connector connects to."""
        return ContactEnergyConnector._UTILITY_NAME



"""App."""

from power_comparison.controller import Controller
from power_comparison.data import Data, Profiles
from power_comparison.view import View


class App:
    """App."""

    def __init__(self) -> None:
        """Initialize App."""
        data = Data()
        profiles = Profiles()
        controller = Controller(data, profiles)
        View(controller)
        data.close()


def main() -> None:
    """Entry point for application."""
    App()


if __name__ == "__main__":
    main()



"""Define the Data class."""
from __future__ import annotations

import sqlite3
from datetime import date, timedelta
from pathlib import Path
from typing import Self

import numpy as np
import numpy.typing as npt

from power_comparison.default_values_utility import DefaultValuesUtility as DVU


class Data:
    """Hold usage data, and manipulation tools."""

    _username: str | None = None
    _user_id: int | None = None

    def __init__(self) -> None:
        """Initialize the Data object without a user.

        To properly initialize with a user, call initialize_user().
        """
        db_filepath = DVU.get_db_file_path()
        DVU.create_dirs(db_filepath)
        self.connection = sqlite3.connect(db_filepath)
        self.cursor = self.connection.cursor()
        self.initialize_database()

    @classmethod
    def from_username(cls, username: str) -> Self:
        """Initialize a Data object from a username."""
        self = cls()
        self.initialize_user(username)
        return self

    def initialize_user(self, username: str) -> None:
        """Ensure the user `username` is initialized."""
        self._username = username
        result = self.cursor.execute(
            "SELECT user_id FROM user_data WHERE username_email=?",
            (self._username,),
        )
        row = result.fetchone()
        if row is not None:
            self._user_id = row[0]
            return
        self.cursor.execute(
            """INSERT INTO user_data (username_email)
        VALUES(?)""",
            (self._username,),
        )
        result = self.cursor.execute(
            "SELECT user_id FROM user_data WHERE username_email=?",
            (self._username,),
        )
        self._user_id = result.fetchone()[0]
        self.connection.commit()

    def initialize_database(self) -> None:
        """Ensure database is initialized and table and user exist."""
        try:
            result = self.cursor.execute(
                "SELECT name FROM sqlite_master WHERE name='usage_data'"
            )
            if result.fetchone() is not None:
                return
            self.cursor.execute(
                """CREATE TABLE
                user_data(
                user_id INTEGER PRIMARY KEY,
                username_email TEXT
                )"""
            )
            self.cursor.execute(
                """CREATE TABLE
                usage_data(
                    user_id INTEGER NOT NULL,
                    date INTEGER NOT NULL, -- Gregorian Ordinal day
                    day INTEGER NOT NULL, -- 0 index day of week
                    hour INTEGER NOT NULL, -- 0 index hour of day
                    value REAL NOT NULL,
                    PRIMARY KEY (user_id, date, hour),
                    FOREIGN KEY (user_id)
                        REFERENCES user_data (user_id)
                )"""
            )
        finally:
            self.connection.commit()

    def get_last_date(self) -> date | None:
        """Return usage data date, or None."""
        if self._user_id is None:
            msg = "Data: _user_id not set"
            raise ValueError(msg)
        result = self.cursor.execute(
            """SELECT date
            FROM usage_data
            WHERE user_id = ?
            ORDER BY date DESC
            """,
            (self._user_id,),
        )
        row = result.fetchone()
        if row is None:
            return None
        return date.fromordinal(row[0])

    def ingest_data(self, data: list[tuple[date, list[float]]]) -> None:
        """Ingest data."""
        if self._user_id is None:
            msg = "Data: _user_id not set"
            raise ValueError(msg)
        for data_date, values in data:
            data_date_ord = data_date.toordinal()
            day = data_date.weekday()
            self.cursor.executemany(
                """INSERT INTO usage_data VALUES(?, ?, ?, ?, ?)""",
                (
                    (self._user_id, data_date_ord, day, hour, value)
                    for hour, value in enumerate(values)
                ),
            )
        self.connection.commit()

    def get_average_usage(
        self, start_date: date | None = None, end_date: date | None = None
    ) -> list[list[float]] | None:
        """Get average of usage data for every hour of every weekday.

        Returns:
            None if there is no or not enough data for the user, else returns
            a list (size seven, ordered by day) of lists
            (size 24, ordered by hour) of floats.

        Args:
            start_date:
                Defaults to one year ago. The first date to include.
            end_date:
                Defaults to today. The last date to include.

        Raises:
            ValueError if initialize_user hasn't been called.
        """
        if self._user_id is None:
            msg = "Data: _user_id not set"
            raise ValueError(msg)
        if end_date is None:
            end_date = date.today()
        if start_date is None:
            start_date = end_date - timedelta(days=365)
        result = self.cursor.execute(
            """SELECT AVG(value), day, hour
            FROM usage_data
            WHERE user_id = ?
            AND date > ?
            AND date < ?
            GROUP BY day, hour
            ORDER BY day ASC, hour ASC""",
            (self._user_id, start_date.toordinal(), end_date.toordinal()),
        )
        data = result.fetchall()
        if len(data) != 7 * 24:
            return None
        return [
            [row[0] for row in data[i * 24 : (i + 1) * 24]] for i in range(7)
        ]

    def get_usage_per_hour(
        self, start_date: date | None = None, end_date: date | None = None
    ) -> list[float] | None:
        """Get average usage data per hour.

        Returns None if there is no data for the user, else returns
        a list of 24 floats ordered by hour.
        Throws ValueError if initialize_user hasn't been called.
        start_date:
            Defaults to one year ago. The first date to include.
        end_date:
        Defaults to today. The last date to include.
        """
        if self._user_id is None:
            msg = "Data: _user_id not set"
            raise ValueError(msg)
        if end_date is None:
            end_date = date.today()
        if start_date is None:
            start_date = end_date - timedelta(days=365)
        result = self.cursor.execute(
            """SELECT AVG(value), hour
            FROM usage_data
            WHERE user_id = ?
            AND date > ?
            AND date < ?
            GROUP BY hour
            ORDER BY hour ASC""",
            (self._user_id, start_date.toordinal(), end_date.toordinal()),
        )
        data = result.fetchall()
        if len(data) == 0:
            return None
        return [row[0] for row in data]

    def close(self) -> None:
        """Close Data."""
        self.connection.commit()
        self.connection.close()


class Profiles:
    """Hold profile data and tools."""

    def __init__(self) -> None:
        """Initialize a Profiles."""

    def get_profile_set_names(self) -> list[str]:
        """Return a list of names of profile sets."""
        return [
            x.name
            for x in Path(DVU.get_profiles_dir()).iterdir()
            if x.is_dir()
        ]

    def _get_profile_data_numpy(
        self, profile: str
    ) -> list[tuple[str, float, npt.NDArray]] | None:
        """Return a list of plan profiles, or None if profile not valid.

        Returns a profile data list of containing tuples of names,
        daily charges, and numpy arrays of shape (7,24).
        """
        path = Path(DVU.get_profiles_dir()) / profile
        if not path.exists():
            return None
        return [
            (
                data_path.stem,
                float(
                    np.loadtxt(
                        str(data_path),
                        dtype=float,
                        delimiter=",",
                        skiprows=8,
                        usecols=1,
                    )
                ),
                np.loadtxt(
                    str(data_path),
                    dtype=float,
                    delimiter=",",
                    skiprows=1,
                    usecols=range(1, 25),
                    max_rows=7,
                ),
            )
            for data_path in path.iterdir()
            if data_path.is_file()
        ]

    def get_profile_data(
        self, profile: str
    ) -> list[tuple[str, float, list[list[float]]]] | None:
        """Return a list of plan profiles or None if profile not valid.

        A plan profile is a tuple of it's name, daily charge, and
        hour usage-based charges, all monetary values in cents.
        A profile data list is nested list of day, followed by hour.
        A profile data list should be size (7,24).
        """
        data = self._get_profile_data_numpy(profile)
        return (
            data
            if data is None
            else [
                (name, daily_charge, numbers.tolist())
                for name, daily_charge, numbers in data
            ]
        )

    def generate_plan_comparison(
        self, usage: list[list[float]], profile: str
    ) -> list[tuple[str, float]] | None:
        """Returns sorted comparison name and cost for year.

        Returns None if profile is not valid.
        """
        profile_data = self._get_profile_data_numpy(profile)
        if profile_data is None:
            return None
        usage_np = np.array(usage, dtype=float)
        result = [
            (
                name,
                (usage_np * data).sum() * ((365 / 100) / 7)
                + daily_charge * (365 / 100),
            )
            for name, daily_charge, data in profile_data
        ]
        result.sort(key=lambda x: x[1])
        return result
